@@ -109,7 +109,7 @@ static void finalize_madfile(value madf)
 {
   madfile_t *mf = Madfile_val(madf);
   if (mf->read_func)
-    caml_remove_global_root(&mf->read_func);
+    caml_remove_generational_global_root(&mf->read_func);
   mad_synth_finish(&mf->synth);
   mad_frame_finish(&mf->frame);
   mad_stream_finish(&mf->stream);
@@ -136,6 +136,7 @@ static inline madfile_t* create_mf()
   mad_frame_init(&mf->frame);
   mad_synth_init(&mf->synth);
   mad_timer_reset(&mf->timer);
+  mf->read_func = (value)NULL;
 
   return(mf);
 }
@@ -161,7 +162,31 @@ CAMLprim value ocaml_mad_openfile(value file)
   CAMLparam1(file);
   CAMLlocal1(block);
   madfile_t *mf;
+
+#ifdef WIN32
+  FILE *fd;
+  const char *fname = String_val(file);
+
+  int size = MultiByteToWideChar(CP_UTF8, 0, fname, -1, NULL, 0);
+  if (size == 0)
+    caml_raise_out_of_memory();
+
+  wchar_t *wfname = malloc(size*sizeof(wchar_t));
+  if (wfname == NULL)
+    caml_raise_out_of_memory();
+
+  if (MultiByteToWideChar(CP_UTF8, 0, fname, -1, wfname, size) == 0)
+    caml_raise_out_of_memory();
+
+  errno_t err = _wfopen_s(&fd, wfname, L"rb");
+  free(wfname);
+
+  if (err != 0)
+    caml_raise_with_arg(*caml_named_value("mad_exn_openfile_error"),
+                        caml_copy_string(strerror(err)));
+#else
   FILE *fd = fopen(String_val(file), "rb");
+#endif
 
   if (!fd)
     caml_raise_with_arg(*caml_named_value("mad_exn_openfile_error"),
@@ -190,7 +215,6 @@ CAMLprim value ocaml_mad_openfile(value file)
     rewind(fd);
 
   mf = create_mf();
-  mf->read_func = (value)NULL;
   mf->fd = fd;
   if (!mf->fd)
     caml_raise_with_arg(*caml_named_value("mad_exn_openfile_error"),
@@ -215,7 +239,7 @@ CAMLprim value ocaml_mad_openstream(value read_func)
 
   mf = create_mf();
   mf->read_func = read_func;
-  caml_register_global_root(&mf->read_func);
+  caml_register_generational_global_root(&mf->read_func);
   mf->fd = 0;
   mf->buf = (unsigned char*)malloc(BUFFER_SIZE);
 
